import powerlaw
from statistics import stdev
import numpy as np

def alpha_estimation(reference_network):
    '''
    Estimate the parameter of power-law ditribution
    '''
    """
    Fit a power law distribution to the degree sequences and
    estimate the parameters of the power-law distribution across networks.

    Parameters:
    - reference_network: A list of adjacency matrices for each network.

    Returns:
    - estimated_alpha : The mean alpha parameter of the power law distribution fitted to the degree sequences.
    - sd_alpha : The standard deviation of the alpha parameters across networks.
    - input_alpha : A random alpha value sampled from a normal distribution centered around estimated_alpha
        with a standard deviation of sd_alpha.
    """
    
    alpha = []
    for current_mat in reference_network:
        degree_seq = current_mat.sum(axis=0)
        degree_seq_non_zero = degree_seq[degree_seq != 0]
        fit = powerlaw.Fit(degree_seq_non_zero, xmin=1, discrete=True)
        alpha.append(fit.power_law.alpha)

    estimated_alpha = round(np.mean(alpha), 2)
    sd_alpha = round(stdev(alpha), 2)
    input_alpha = np.random.normal(loc=estimated_alpha, scale=sd_alpha, size=1)
    
    return estimated_alpha, sd_alpha, input_alpha



from .pns_construction import *
from .KGML_to_Matrix import *
from .Network_score_with_cor import *
from .show_graph_with_labels import *
from .import_references import *
from .alpha_estimation import *
from .deviation_calculation import *
from .generate_ensemble_network import *



import numpy as np
import pandas as pd

def import_references(pathway_ids, base_path):
    """
    This function can read  output of KGML_to_Matrix to builds undirected adjacency matrices for each pathway specified by pathway_ids,

    Parameters:
    - pathway_ids : The IDs of the pathways for which to construct networks.
    - base_path : The base directory path where the reference networks are stored.

    Returns:
    - reference_network : A list containing the undirected adjacency matrices for each pathway.
    - network_size : A list containing the size of each network.
    - reference_sparsity : A list containing the sparsity of each network.
    """

    reference_network = []
    network_size = []
    reference_sparsity = []
    
    for pathway_id in pathway_ids:
        file_name = f"{base_path}/hsa{pathway_id}(directed)"
        directed_adjmatrix = pd.read_pickle(file_name).to_numpy()
        undirected_adjmatrix = directed_adjmatrix + directed_adjmatrix.T
        np.fill_diagonal(undirected_adjmatrix, 0)
        undirected_adjmatrix[undirected_adjmatrix > 1] = 1
        reference_network.append(undirected_adjmatrix)

        one_counts = np.sum(undirected_adjmatrix[np.tril_indices_from(undirected_adjmatrix, k=-1)])
        size = undirected_adjmatrix.shape[0]
        network_size.append(size)
        sparsity = 2 * one_counts / (size * (size - 1))
        reference_sparsity.append(sparsity)

    return reference_network, network_size, reference_sparsity



import numpy as np
import networkx as nx

def calculate_network_properties(networks):
    ave_path_len = []
    max_degree_centrality = []
    transitivity = []

    for net in networks:
        G = nx.from_numpy_array(np.array(net))
        # Average path length
        dis_mat = nx.floyd_warshall_numpy(G)
        dis_mat_without_inf = np.where(np.isinf(dis_mat), len(dis_mat), dis_mat)
        ave_path_length = dis_mat_without_inf.sum()/len(dis_mat)/len(dis_mat)
        ave_path_len.append(ave_path_length)
        # Max degree centrality
        max_degree_centrality.append(max(nx.degree_centrality(G).values()))
        # Transitivity
        transitivity.append(nx.transitivity(G))

    return ave_path_len, max_degree_centrality, transitivity


def calculate_means(values):
    return np.mean(values)


def standardize_deviation(actual, mean):
    deviation = (actual - mean) / np.std(actual)
    deviation[np.isnan(deviation)] = 0
    return deviation



import numpy as np
import networkx as nx
import random



def Network_score_with_cor(Candidate_network, Structure_score, Confidence_mat, Gene_list, Network_ranking_size, Permute_times = 10000, Permute_size = 2, Update_count = 1000):
    """
    Calculates network scores based on network structure and correlation, applying permutations to get optimal gene labels.

    Parameters:
    - Candidate_network : potential network structures.
    - Structure_score : Deviation measures for the structures.
    - Confidence_mat : confidence matrix of data.
    - Gene_list : genes of interest.
    - Network_ranking_size : top n% potential networks for further analysis.
    - Permute_times : maximum permutation times for each structure.
    - Permute_size : number of genes to switch in each permutation.
    - Update_count : threshold for increasing permutation size.

    Returns:
    - A tuple containing various outputs related to the network scoring and permutation process.
    """
    tmp_structure_score = Structure_score.copy()
    tmp_structure_score.sort()
    final_gene_label = []
    final_net_score = []
    final_structure_list = []
    Graph_list = []
    unique_network_count = 0
    structure_count = 0
    structure_count_list = []
    conv_net_score = []
    update_net_score = []
    Permute_size_list = []
    Permute_gene_label = []
    
    # execution
    while unique_network_count < Network_ranking_size and structure_count < len(tmp_structure_score):
        net_index = np.where(Structure_score == tmp_structure_score[structure_count])[0][0]
        structure_count += 1
        current_net = np.array(Candidate_network[net_index])

        isomorphic_result = False
        if unique_network_count != 0:
            current_graph = nx.from_numpy_array(current_net)
            for iso_test in range(unique_network_count)[::-1]:
                isomorphic_result = nx.is_isomorphic(Graph_list[iso_test], current_graph)
                if isomorphic_result == True:
                    break

        if isomorphic_result == False:
            structure_count_list.append(structure_count)
            Graph_list.append(nx.from_numpy_array(current_net))
            final_structure_list.append(current_net)
            unique_network_count += 1
            current_degree_seq = sum(np.array(current_net))
            max_degree_node_index = np.where(current_degree_seq == max(current_degree_seq))[0][0]
            cor_sum = np.sum(Confidence_mat, axis=0)
            max_cor_sum_gene = Gene_list[np.where(cor_sum == max(cor_sum))[0][0]]

            gene_label = np.array(Gene_list.copy())
            permute_candidate = np.array(np.where(gene_label == max_cor_sum_gene)[0][0])
            permute_candidate = np.append(permute_candidate, max_degree_node_index)
            gene_label[permute_candidate] = gene_label[permute_candidate[::-1]]
            gene_cor = np.copy(Confidence_mat)
            gene_cor[:,permute_candidate] = gene_cor[:,permute_candidate[::-1]]
            gene_cor[permute_candidate,:] = gene_cor[permute_candidate[::-1],:]
            net_score = 0
            for jj in range(len(Confidence_mat)):
                net_score = net_score + sum(gene_cor[jj]*current_net[jj])

            tmp_conv_net_score = []
            tmp_Permute_size_list = []
            tmp_update_net_score = []
            tmp_Permute_gene_label = []

            no_update_count = 0
            total_iterations = 0
            tmp_permute_size = Permute_size
            while total_iterations < Permute_times and tmp_permute_size < len(Gene_list):
                tmp_gene_label = np.copy(gene_label)
                permute_candidate = random.sample(range(len(Gene_list)), tmp_permute_size)
                tmp_Permute_size_list.append(tmp_permute_size)
                after_permute = permute_candidate.copy()
                while after_permute == permute_candidate:
                    after_permute = random.sample(permute_candidate, tmp_permute_size)
                tmp_gene_label[permute_candidate] = tmp_gene_label[after_permute]
                tmp_gene_cor = np.copy(gene_cor)
                tmp_gene_cor[:, permute_candidate] = tmp_gene_cor[:, after_permute]
                tmp_gene_cor[permute_candidate,:] = tmp_gene_cor[after_permute,:]

                tmp_score = 0
                for jj in range(len(Confidence_mat)):
                    tmp_score = tmp_score + sum(tmp_gene_cor[jj]*current_net[jj])
                if (tmp_score > net_score):
                    gene_label = tmp_gene_label.copy()
                    net_score = tmp_score.copy()
                    no_update_count = 0
                    tmp_Permute_gene_label.append(gene_label)
                else:
                    no_update_count += 1

                tmp_conv_net_score.append(tmp_score)
                tmp_update_net_score.append(net_score)

                if no_update_count >= Update_count:
                    tmp_permute_size += 1
                    no_update_count = 0
                    i = 0

                total_iterations += 1

            conv_net_score.append(tmp_conv_net_score)
            update_net_score.append(tmp_update_net_score)
            Permute_gene_label.append(tmp_Permute_gene_label)
            final_gene_label.append(gene_label)
            final_net_score.append(net_score)
            Permute_size_list.append(tmp_Permute_size_list)
    
    # Sort by network score
    sorted_index = np.argsort(final_net_score)
    sorted_index = sorted_index[::-1]
    sorted_structure_output = []
    sorted_score_output = []
    sorted_gene_label_output = []

    for i in sorted_index:
        sorted_structure_output.append(final_structure_list[i])
        sorted_score_output.append(final_net_score[i])
        sorted_gene_label_output.append(final_gene_label[i])

    return sorted_structure_output, sorted_gene_label_output, sorted_score_output
    




import numpy as np
import networkx as nx
import matplotlib.pyplot as plt


def show_graph_with_labels(adjacency_matrix, input_label=""):
    """
    Displays a graph (network) represented by an adjacency matrix with optional labels.

    Parameters:
    - adjacency_matrix : the network structure.
    - input_label : labels for the nodes of the graph (network).
    """
    gr = nx.from_numpy_array(np.array(adjacency_matrix))
    graph_pos = nx.spring_layout(gr, k=0.50, iterations=50)

    nx.draw_networkx_nodes(gr, graph_pos, node_color='#1f78b4', node_size=220, alpha=0.6)
    nx.draw_networkx_edges(gr, graph_pos, width=2, alpha=0.3)
    if input_label:
        labels = {i: str(label) for i, label in enumerate(input_label)}
        nx.draw_networkx_labels(gr, graph_pos, labels)
    else:
        nx.draw_networkx_labels(gr, graph_pos)
    plt.show()


import setuptools


setuptools.setup(
    name='NEKBDD',
    version='0.1.0',
    packages=setuptools.find_packages(),
    description='A package for NEKBDD: a nonparametric ensemble knowledge-based data-driven genetic network construction',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Chen-Po Liao',
    author_email='liaochenpo@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.6',
)



import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

def generate_ensemble_network(Structure_output, Gene_label_output, Real_data_genes, reference_sparsity):
    """
    Generates an ensemble network based on the sorted structure outputs and gene labels.
    
    Parameters:
    - Structure_output : List of network structures.
    - Gene_label_output : List of gene labels for each network.
    - Real_data_genes : List of all genes considered in the analysis.
    - reference_sparsity : reference sparsity values to compare against.
        
    Returns:
    - ensemble_network : NetworkX graph. The generated ensemble network based on the optimal threshold.
    - ensemble_structure: The network after thresholding
    - potential_proportion: The networks before threholding
    """
    
    thresholds = np.arange(0, 1.05, 0.05)

    
    thresholds = np.arange(0, 1.05, 0.05)
    Structure_shuffle_output = [None] * len(Structure_output)

    for i in range(len(Structure_output)):
        match_index = []
        tmp_gene_label = np.array(Gene_label_output[i])
        for j in range(len(Real_data_genes)):
            match_index.append(np.where(tmp_gene_label == Real_data_genes[j])[0][0])
        shuffle_structure = np.copy(Structure_output[i])
        shuffle_structure = shuffle_structure[:, match_index]
        shuffle_structure = shuffle_structure[match_index,:]
        Structure_shuffle_output[i] = shuffle_structure

    potential_proportion = np.sum(Structure_shuffle_output, axis=0)/len(Structure_shuffle_output)

    ensemble_sparsity = [
        np.count_nonzero(np.where(potential_proportion > threshold, 1, 0)) /
        (((np.where(potential_proportion > threshold, 1, 0).size ** 0.5) * ((np.where(potential_proportion > threshold, 1, 0).size ** 0.5) - 1)))
        for threshold in thresholds
    ]

    closest_index, min_distance = None, float('inf')
    for index, sparsity in enumerate(ensemble_sparsity):
        distance = abs(sparsity - np.mean(reference_sparsity))
        if distance < min_distance:
            min_distance = distance
            closest_index = index
    print("Closest threshold: {}".format(thresholds[closest_index]))
    print("Closest sparsity: {}".format(ensemble_sparsity[closest_index]))

    threshold = round(thresholds[closest_index], 2)
    ensemble_structure = np.where(potential_proportion > threshold, 1, 0)
    
    ensemble_network = nx.from_numpy_array(ensemble_structure)
    mapping = {i: Real_data_genes[i] for i in range(len(Real_data_genes))}
    ensemble_network = nx.relabel_nodes(ensemble_network, mapping)

    return ensemble_structure, ensemble_network, potential_proportion



import xml.etree.ElementTree as ET
import pandas as pd
import numpy as np
import itertools


def KGML_to_Matrix(pathway_name, KGML_file_path = "", save_path = ""):
    """
    Converts KGML file to adjacency matrix and node detail for KEGG human biological pathway analysis.

    Parameters:
    - pathway_name : the name of the pathway.
    - KGML_file_path : the file path of the KGML file.
    - save_path : the directory path where the output files will be saved.

    The function extracts node information and relationships from the KGML file,
    processes them, and outputs an adjacency matrix and node details as pandas DataFrames.
    """

    tree = ET.parse(KGML_file_path)
    root = tree.getroot()

    data_raw = []
    for child in root:
        data = {child.tag: child.attrib, "graphics": [], "component": [], "subtype": []}
        for children in child:
            data[children.tag].append(children.attrib)
        data_raw.append(data)

    for ii in range(len(data_raw)):
        data_raw[ii] = {k: v for k, v in data_raw[ii].items() if v}

    data_entry_raw, data_relation_raw = [], []
    for item in data_raw:
        if "entry" in item:
            data_entry_raw.append(item)
        if "relation" in item:
            data_relation_raw.append(item)

    for item in data_entry_raw:
        graphics = item.get("graphics", [{}])[0]
        if "name" in graphics:
            graphics["node_name"] = graphics.pop("name")
        if "type" in graphics:
            graphics["type1"] = graphics.pop("type")
    data_entry = [{**item["entry"], **item["graphics"][0]} for item in data_entry_raw]
    data_entry = pd.DataFrame(data_entry)[["id", "name", "type", "node_name"]]

    number = []
    for ii in range(len(data_entry_raw)):
        if "component" in data_entry_raw[ii]:
            number.append(len(data_entry_raw[ii]["component"]))
        else:
            number.append(int(1))
    component_id = []
    for ii in range(len(number)):
        for jj in range(number[ii]):
            component_id.append(range(len(number))[ii])

    data_entry_component = pd.DataFrame(data_entry.iloc[component_id ,:])

    component = []
    for ii in range(len(data_entry_raw)):
        if "component" not in data_entry_raw[ii]:
            component.append(float("NaN"))
        if "component" in data_entry_raw[ii]:
            for jj in data_entry_raw[ii]["component"]:
                component.append(jj["id"])

    component = pd.DataFrame(component, index = component_id)
    component.columns = ["component"]
    data_entry_new = pd.concat([data_entry_component, component], axis = 1)

    data_entry = data_entry[data_entry["type"] != "map"].reset_index(drop=True)

    unique_correspond = None
    unique_entryname = None
    unique_entryid = None

    unique_entryname = data_entry["name"].unique()
    unique_erntryid = data_entry["id"][data_entry["name"].duplicated() == False]

    match = lambda a, b: [ b.index(x) if x in b else None for x in a ]
    position_repeat = match(list(data_entry["name"][data_entry["name"].duplicated()]), list(data_entry["name"]))

    Var1 = list(data_entry["id"][data_entry["name"].duplicated()])
    Var2 = list(data_entry["id"][position_repeat])
    unique_correspond = {"Var1":Var1, "Var2":Var2}
    unique_correspond = pd.DataFrame(unique_correspond)

    data_relation = pd.DataFrame([item["relation"] for item in data_relation_raw])[["entry1", "entry2"]]

    if len(data_relation) == 0:
        return print("There are no relation in the ", pathway_name, "!")
    else:
        separate_from = list(data_entry_new[data_entry_new["type"] == "group"]["id"])
        separate_to = list(data_entry_new["component"][data_entry_new["component"].isnull() == False])

        relation_new = []
        for ii in range(len(data_relation.index)):
            if [i in data_relation.iloc[ii,].iloc[0] for i in separate_from].count(True) > 0:
                relation1 = [separate_to[i] for i, v in enumerate([i in data_relation.iloc[ii,].iloc[0] for i in separate_from]) if v == True]
            else:
                relation1 = [data_relation.iloc[ii,].iloc[0]]
            if [i in data_relation.iloc[ii, 1] for i in separate_from].count(True) > 0:
                relation2 = [separate_to[i] for i, v in enumerate([i in data_relation.iloc[ii, 1] for i in separate_from]) if v == True]

            else:
                relation2 = [data_relation.iloc[ii,].iloc[1]]
            relation_new.append([(x, y) for x in relation1 for y in relation2])
        relation_new = pd.concat(list(map(pd.DataFrame, relation_new)))
        relation_new.columns = ["Var1", "Var2"]
        relation_new = relation_new.reset_index(drop = True)

        def relation_group_fn(xx):
            sub_group = [separate_to[i] for i, v in enumerate([i in xx for i in separate_from]) if v == True]
            return pd.DataFrame(list(itertools.combinations(sub_group, 2)))

        if len(separate_from) > 0:
            relation_group = pd.concat(list(map(pd.DataFrame, list(map(relation_group_fn, sorted(list(set(separate_from))))))))
            relation_group.columns = ["Var1", "Var2"]
            relation_group = relation_group.reset_index(drop = True)
            Reverse_relation = relation_group[["Var2", "Var1"]]
            Reverse_relation.columns = ["Var1", "Var2"]
            relation_group = pd.concat([relation_group, Reverse_relation]).reset_index(drop = True)
            relation_new = pd.concat([relation_new, relation_group]).reset_index(drop = True)

        relationship = pd.DataFrame(np.zeros((len(data_entry["id"]), len(data_entry["id"])), dtype = int), index = list(data_entry["id"]), columns = list(data_entry["id"]))
        position1 = match(relation_new["Var1"], list(data_entry["id"]))
        position2 = match(relation_new["Var2"], list(data_entry["id"]))
        relation_position = pd.DataFrame({"position1" :position1, "position2" :position2})
        for ii in range(len(relation_position.index)):
            x = relation_position.iloc[ii, 0]
            y = relation_position.iloc[ii, 1]
            relationship.iloc[x, y] = 1

        pos1 = match(unique_correspond["Var1"], list(data_entry["id"]))
        pos2 = match(unique_correspond["Var2"], list(data_entry["id"]))

        if len(pos1) != 0:
            for ii in range(len(pos1)):
                relationship.iloc[pos2[ii], ] = relationship.iloc[pos2[ii], ] + relationship.iloc[pos1[ii], ]
                relationship.iloc[:, pos2[ii]] = relationship.iloc[:, pos2[ii]] + relationship.iloc[:, pos1[ii]]
        if len(pos1) > 0:
            pos1_relationship = [relationship.columns[ii] for ii in pos1]
            relationship = relationship.drop(pos1_relationship, axis = 1)
            relationship = relationship.drop(pos1_relationship, axis = 0)

            data_entry.index = list(data_entry["id"])
            entry_pos1 = [data_entry["id"].iloc[ii] for ii in pos1]
            node_detail = data_entry.drop(entry_pos1, axis = 0)
            node_detail = node_detail[["name", "type", "node_name"]]

            if list(data_entry["type"] == "group").count(True) != 0:
                delete_group = data_entry.drop(entry_pos1, axis = 0)
                delete_group_drop = [delete_group.index[ii] for ii, vv in enumerate(delete_group["type"] == "group") if vv == True]
                relationship = relationship.drop(delete_group_drop, axis = 0)
                relationship = relationship.drop(delete_group_drop, axis = 1)
                node_detail = node_detail.drop(delete_group_drop, axis = 0)
        else:
            data_entry.index = list(data_entry["id"])
            node_detail = data_entry[["name", "type", "node_name"]]

        for ii in range(len(relationship.index)):
            for jj in range(len(relationship.columns)):
                if relationship.iloc[ii, jj] > 1:
                    relationship.iloc[ii, jj] = 1
        adj_matrix = relationship

        file_name = save_path + pathway_name + "(directed)"
        adj_matrix.to_pickle(file_name)
        file_name = save_path + pathway_name + "(node_detail)"
        node_detail.to_pickle(file_name)
        return "Success"



import numpy as np
from itertools import combinations
import math
from scipy import stats

def EGtest(DS_input):
    """
    Tests if a given degree sequence DS_input can form a simple graph according to the Erdős–Gallai theorem.

    Parameters:
    - DS_input : the degree sequence of a graph.

    Returns:
    - 'success' if the sequence can form a simple graph, 'failure' otherwise.
    """
    if np.sum(DS_input) % 2 == 1 or np.sum(DS_input) > len(DS_input) * (len(DS_input) - 1):
        return 'failure'
    if np.sum(DS_input) != 0 and 0 in DS_input:
        DS_input = np.delete(DS_input, np.where(DS_input == 0))

    DS_input = np.sort(DS_input)[::-1]
    for index in range(len(DS_input)):
        k = index + 1
        tmp_DS_input = np.array([
            k if DS_input[x] > k and x >= k else DS_input[x] for x in range(len(DS_input))
        ])
        if sum(tmp_DS_input[0:k]) > k * (k - 1) + np.sum(tmp_DS_input[k:]):
            return 'failure'

    return 'success'


def rightmost_adj(DS_input):
    """
    Identifies the rightmost adjacency set for a graph based on its degree sequence.

    Parameters:
    - DS_input : the degree sequence of a graph.

    Returns:
    - rightmost_adj_set : indices of the rightmost adjacency set.
    """

    non_zero_degree_node_index = np.array([i for i, x in enumerate(DS_input) if x > 0])

    if 0 in DS_input:
        DS_input = np.delete(DS_input, np.where(DS_input == 0))

    order_index = np.argsort(-DS_input)
    DS_input = np.sort(DS_input)[::-1]

    rightmost_adj_set = np.array([], dtype=int)
    for non_leading_node in range(1, len(DS_input))[::-1]:
        tmp_DS = np.copy(DS_input)
        tmp_DS[0] -= 1
        tmp_DS[non_leading_node] -= 1

        if tmp_DS[0] != 0:
            DS_for_test = np.array([tmp_DS[i] - 1 if i <= tmp_DS[0] else tmp_DS[i] for i in range(1, len(tmp_DS))])
        else:
            DS_for_test = np.copy(tmp_DS)

        if EGtest(DS_for_test) == 'success':
            rightmost_adj_set = np.append(rightmost_adj_set, non_leading_node)
            if tmp_DS[0] == 0:
                break
            DS_input = np.copy(tmp_DS)

    rightmost_adj_set = np.array(non_zero_degree_node_index[order_index[rightmost_adj_set]])
    return rightmost_adj_set



def left_adj(current_adj_mat, DS_input_original, DS_input_current, rightmost_adj, leading_node_index):
    """
    Determines the left adjacency set for the current graph.

    Parameters:
    - current_adj_mat : the current adjacency matrix.
    - DS_input_original : the original degree sequence.
    - DS_input_current : the current degree sequence after updates.
    - rightmost_adj : indices of the rightmost adjacency set.
    - leading_node_index : index of the leading node.

    Returns:
    - left_adj_set : indices of nodes forming the left adjacency set.
    """

    leading_node_degree = DS_input_current[leading_node_index]
    tmp_DS = np.copy(DS_input_current)
    tmp_DS[leading_node_index] = 0
    order_index = np.argsort(-tmp_DS)
    tmp_DS = np.sort(tmp_DS)[::-1]

    non_zero_DS = np.delete(tmp_DS, np.where(tmp_DS == 0))
    number_of_min = len(np.where(non_zero_DS == min(non_zero_DS))[0])
    non_min_node_index = np.where(non_zero_DS != min(non_zero_DS))[0]
    number_of_non_min = len(non_min_node_index)

    tmp_left_adj_set = [[0]*leading_node_degree]
    i_start = max(leading_node_degree - number_of_min, 0)
    i_end = max(min(number_of_non_min, leading_node_degree), 0) + 1

    duplicate_marker = [mat.copy() for mat in current_adj_mat]
    [duplicate_marker[i].append(DS_input_original[i]) for i in range(len(DS_input_original))]
    duplicated_index = [0]*len(duplicate_marker)
    for i in range(len(duplicate_marker)):
        for j in range(len(duplicate_marker)):
            if duplicate_marker[j] == duplicate_marker[i]:
                duplicated_index[i] = j
    duplicated_index = [duplicated_index[order_index[i]] for i in range(len(order_index))]
    for i in range(i_start, i_end):
        if i == 1 and number_of_non_min == 1:
            first_part = [list(non_min_node_index)]
        elif i != 0:
            first_part = [list(l) for l in combinations(non_min_node_index,i)]
            duplicated_mat = [0]*len(first_part)
            for j in range(len(first_part)):
                duplicated_mat[j] = [duplicated_index[first_part[j][k]] for k in range(len(first_part[j]))]
            unique_index = []
            unique_value = []
            for j in range(len(duplicated_mat))[::-1]:
                x = duplicated_mat[j]
                if x not in unique_value:
                    unique_value.append(x)
                    unique_index.append(j)
            first_part = [first_part[m] for m in range(len(first_part)) if m in unique_index]

        if i != leading_node_degree:
            min_degree_node = np.where(non_zero_DS == min(non_zero_DS))[0]
            if len(min_degree_node) == 1:
                second_part = [list(min_degree_node)]
            else:
                second_part = [list(l) for l in combinations(min_degree_node,leading_node_degree-i)]

            duplicated_mat = [0]*len(second_part)
            for j in range(len(second_part)):
                duplicated_mat[j] = [duplicated_index[second_part[j][k]] for k in range(len(second_part[j]))]
            unique_index = []
            unique_value = []
            for j in range(len(duplicated_mat))[::-1]:
                x = duplicated_mat[j]
                if x not in unique_value:
                    unique_value.append(x)
                    unique_index.append(j)
            second_part = [second_part[m] for m in range(len(second_part)) if m in unique_index]

        if i == 0:
            combine_two_part = second_part
        elif i == leading_node_degree:
            combine_two_part = first_part
        else:
            combine_two_part = [x + y for x in first_part for y in second_part]
        tmp_left_adj_set = tmp_left_adj_set + combine_two_part
    tmp_left_adj_set.remove(tmp_left_adj_set[0])

    mapping_index = [np.where(tmp_DS == tmp_DS[k])[0][0] for k in range(len(tmp_DS))]
    colex_order_rightmost = [np.where(order_index == rightmost_adj[i])[0][0] for i in range(len(rightmost_adj))]
    colex_score_rightmost = np.sum([2**(mapping_index[colex_order_rightmost[i]]) for i in range(len(colex_order_rightmost))])
    colex_score_left = []
    for i in range(len(tmp_left_adj_set)):
        colex_score_i = np.sum([2**(mapping_index[tmp_left_adj_set[i][j]]) for j in range(len(colex_order_rightmost))])
        colex_score_left.append(colex_score_i)
    check_to_the_left = np.array(colex_score_left <= colex_score_rightmost, dtype=bool)

    tmp_left_adj_set = np.asarray(tmp_left_adj_set)
    left_adj_set = tmp_left_adj_set[check_to_the_left]
    left_adj_set = [list(order_index[k]) for k in left_adj_set]
    return left_adj_set


def connect_adj_set(leading_node_index, current_adj_mat, adj_set):
    """
    Connects nodes in the adjacency set to the leading node, updating the adjacency matrix.

    Parameters:
    - leading_node_index : index of the leading node in the adjacency matrix.
    - current_adj_mat : current adjacency matrix of the graph.
    - adj_set : set of nodes to be connected to the leading node.

    Returns:
    - output_mat : updated adjacency matrix with new connections.
    """

    output_mat = []
    for ii in range(len(adj_set)):
        tmp_mat = [row.copy() for row in current_adj_mat]
        for jj in range(len(adj_set[0])):
            tmp_mat[leading_node_index][adj_set[ii][jj]] = tmp_mat[adj_set[ii][jj]][leading_node_index] = 1
        output_mat.append([row.copy() for row in tmp_mat])
    return output_mat


def net_gen(original_DS, max_pns=1000):
    """
    Generates potential network structures from a given degree sequence.

    Parameters:
    - original_DS : the original degree sequence for network generation.
    - max_pns: the maximum for generating potential network structures per degree sequence.

    Returns:
    - complete_adj_mat : generated potential network structures.
    """

    sum_DS = np.sum(original_DS)
    rows, cols = len(original_DS), len(original_DS)
    incomplete_adj_mat = [[[0] * cols for _ in range(rows)]]
    complete_adj_mat = []

    while (len(incomplete_adj_mat) != 0) and (len(complete_adj_mat) < max_pns):
        last_matrix = incomplete_adj_mat.pop()
        current_DS = original_DS - np.array([sum(row) for row in last_matrix])

        if np.sum(current_DS != 0) > 1:
            leading_node = np.argmax(current_DS)
            rightmost_adj_set = rightmost_adj(current_DS)

            left_adj_set = left_adj(
                current_adj_mat=last_matrix,
                DS_input_original=original_DS,
                DS_input_current=current_DS,
                rightmost_adj=rightmost_adj_set,
                leading_node_index=leading_node
            )

            new_matrices = connect_adj_set(
                leading_node_index=leading_node,
                current_adj_mat=last_matrix,
                adj_set=left_adj_set
            )

            for matrix in new_matrices:
                if sum([sum(row) for row in matrix]) == sum_DS:
                    complete_adj_mat.append(matrix)
                else:
                    incomplete_adj_mat.append(matrix)

    return complete_adj_mat


def truncated_power_law(alpha, maximum_value):
    """
    Generates a discrete truncated power law distribution.

    Parameters:
    - alpha : the parameter of a power-law distribution.
    - maximum_value : the maximum value of the distribution.

    Returns:
    - A sample of the distribution.
    """
    x = np.arange(1, maximum_value + 1, dtype='float')
    pmf = 1 / x**alpha
    pmf /= pmf.sum()
    return stats.rv_discrete(values=(range(1, maximum_value + 1), pmf))


def pns_generation(Real_data_genes, input_alpha, DS_number=10):
    """
    Generates potential network structure based on degree sequences drawn from truncated power law distribution
    
    Parameters:
    - Real_data_genes: List of real data genes to determine the size of potential network structures.
    - input_alpha: The alpha parameter for the power law distribution.
    - DS_number: totoal number of degree sequences

    Returns:
    - potential_net: List of generated potential network structures.
    - sample_DS_data: List of sampled degree sequences that passed EG test.
    """
    
    potential_net_size = len(Real_data_genes)
    potential_net = []
    sample_DS_data = []

    for DS_count in range(DS_number):
        EG_result = "failure"
        while EG_result == "failure":
            d = truncated_power_law(alpha=input_alpha, maximum_value=potential_net_size - 1)
            sample_DS = d.rvs(size=potential_net_size)
            sample_DS.sort()
            sample_DS = sample_DS[::-1]
            EG_result = EGtest(sample_DS)

        sample_net = net_gen(np.array(sample_DS))
        sample_DS_data.append(sample_DS)
        for net_index in range(len(sample_net)):
            potential_net.append(sample_net[net_index])

    print("Total potential network structures:", len(potential_net))
    
    return potential_net, sample_DS_data

